// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "### Top-level Clauses\n"
"\n"
"    (version 1)\n"
"\n"
"    (rule <rule_name> <rule_clause> ...)\n"
"\n"
"<br>\n"
"\n"
"### Rule Clauses\n"
"\n"
"    (constraint <constraint_type> ...)\n"
"\n"
"    (condition \"<expression>\")\n"
"\n"
"    (layer \"<layer_name>\")\n"
"\n"
"    (severity <severity_name>)\n"
"\n"
"\n"
"<br>\n"
"\n"
"" );
