/*
 * This program source code file is part of KiCad, a free EDA CAD application.
 *
 * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you may find one here:
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 * or you may search the http://www.gnu.org website for the version 2 license,
 * or you may write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#ifndef ALTIUM_PROJECT_VARIANTS_H
#define ALTIUM_PROJECT_VARIANTS_H

#include <map>
#include <vector>
#include <wx/string.h>

/**
 * A single component variation within an Altium project variant.
 *
 * In Altium, each variant can override individual components. Kind=1 means
 * "Not Fitted" (DNP), Kind=0 means "Alternate" with optional field overrides
 * supplied via the AlternatePart string.
 */
struct ALTIUM_VARIANT_ENTRY
{
    wxString                         designator;
    wxString                         uniqueId;
    int                              kind = 0;
    std::map<wxString, wxString>     alternateFields;
};


/**
 * A project-level assembly variant parsed from an Altium .PrjPcb file.
 *
 * Each [ProjectVariantN] section in the .PrjPcb becomes one of these.
 */
struct ALTIUM_PROJECT_VARIANT
{
    wxString                                 name;
    wxString                                 description;
    std::vector<ALTIUM_VARIANT_ENTRY>        variations;
};


/**
 * Parse all [ProjectVariantN] sections from an Altium .PrjPcb project file.
 *
 * @param aPrjPcbPath Full path to the .PrjPcb file.
 * @return Vector of project variants with their per-component entries.
 */
std::vector<ALTIUM_PROJECT_VARIANT> ParseAltiumProjectVariants( const wxString& aPrjPcbPath );

#endif // ALTIUM_PROJECT_VARIANTS_H
