// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "### Documentation\n"
"\n"
"For the full documentation see [https://docs.kicad.org](https://go.kicad.org/docs/GetMajorMinorVersion/en/pcbnew/pcbnew.html#custom_design_rules)." );
