#!/bin/bash
sudo apt-get update && sudo apt-get install -y \
  debhelper \
  cmake \
  cmake-curses-gui \
  build-essential \
  pkg-config \
  grep \
  doxygen \
  dblatex \
  po4a \
  asciidoc \
  source-highlight \
  libcairo2-dev \
  libglu1-mesa-dev \
  libgl1-mesa-dev \
  libx11-dev \
  mesa-common-dev \
  libgtk-3-dev \
  libglm-dev \
  libwxgtk3.2-dev \
  libwxgtk-webview3.2-dev \
  python3-wxgtk4.0 \
  libbz2-dev \
  libssl-dev \
  libzstd-dev \
  zlib1g-dev \
  python3-dev \
  swig \
  python3-pytest \
  python3-cairosvg \
  python3-numpy \
  libocct-modeling-algorithms-dev \
  libocct-modeling-data-dev \
  libocct-data-exchange-dev \
  libocct-visualization-dev \
  libocct-foundation-dev \
  libocct-ocaf-dev \
  libgit2-dev \
  libsecret-1-dev \
  libboost-all-dev \
  libcurl4-openssl-dev \
  libngspice0-dev \
  unixodbc-dev \
  libspnav-dev \
  libnng-dev \
  libprotobuf-dev \
  protobuf-compiler \
  shared-mime-info \
  libzint-dev \
  libpoppler-dev \
  libpoppler-glib-dev
