// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "- This calculator finds combinations of standard E-series (between 10Ω and 1MΩ) to create arbitrary values.\n"
"- You can enter the required resistance from 0.0025 to 4000 kΩ.\n"
"- Solutions using up to 4 components are given.\n"
"\n"
"The requested value is always excluded from the solution set.<br>\n"
"Optionally up to two additional values can be excluded in case of component availability problems.\n"
"\n"
"Solutions are given in the following formats:\n"
"\n"
"	R1 + R2 +...+ Rn	resistors in series\n"
"	R1 | R2 |...| Rn	resistors in parallel\n"
"	R1 + (R2|R3)...		any combination of the above\n"
"" );
