// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "You can use this calculator to check if a small track can handle a large current for a short period of time.<br>\n"
"This tool allows you to design a track fuse but should be used as an estimate only.\n"
"\n"
"The calculator estimates the energy required to heat the wire up<br>\n"
"to its melting point as well as the energy required for the change of phase.<br>\n"
"This energy is then compared to the one dissipated by the wire resistance." );
