// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "Passive components are commonly made with E-series values appropriate to their precision.\n"
"Capacitors commonly use E12 values. 10% and 5% resistors commonly use E24 values. 1%\n"
"resistors use E96 values. Other series are not commonly used.\n"
"\n"
"To select a value begin with the calculated target value and then round it to 2\n"
"significant figures for E24 or below or 3 significant figures for E48 and up.\n"
"Then find the value in the table which is nearest to the significant figures\n"
"remaining and substitute it for those figures.\n"
"\n"
"For example if the calculated target value is 16,834.2Ω then this rounds to 16,800Ω.\n"
"The nearest value to 168 is 169 and the selected E96 value is 16.9kΩ.\n"
"\n"
"The value 0 is a special case and is not present in any series." );
