// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "### Pi Attenuator\n"
"___a___ is attenuation in dB<br>\n"
"___Z<sub>in</sub>___ is desired input impedance in &#x2126;<br>\n"
"___Z<sub>out</sub>___ is desired output impedance in &#x2126;<br>\n"
"\n"
"___K = V<sub>I</sub>/V<sub>O</sub> = 10<sup>a/20</sup>___<br>\n"
"___L = K<sup>2</sup> = 10<sup>a/10</sup>___<br>\n"
"___A = (L+1) / (L&minus;1)___<br><br>\n"
"___R2 = (L&minus;1) / 2&middot;&radic;(Z<sub>in</sub> &middot; Z<sub>out</sub> / L)___<br>\n"
"___R1 = 1 / (A/Z<sub>in</sub> &minus; 1/R2)___<br>\n"
"___R3 = 1 / (A/Z<sub>out</sub> &minus; 1/R2)___\n"
"" );
