/*
 * This program source code file is part of KiCad, a free EDA CAD application.
 *
 * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you may find one here:
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 * or you may search the http://www.gnu.org website for the version 2 license,
 * or you may write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#pragma once

// App icons, we list these first, there will only be one per-exe but they must be first
// out of all possible icons
#define IDI_APP_KICAD_ICON 100
#define IDI_APP_EESCHEMA_ICON 101
#define IDI_APP_PCBNEW_ICON 102
#define IDI_APP_PLEDITOR_ICON 103
#define IDI_APP_BITMAP2COMPONENT_ICON 104
#define IDI_APP_PCBCALC_ICON 105
#define IDI_APP_GERBVIEW_ICON 106

// File icons
#define IDI_FILE_PROJECT_ICON 200
#define IDI_FILE_SCHEMATIC_ICON 201
#define IDI_FILE_SYMBOL_ICON 202
#define IDI_FILE_PCB_ICON 203
#define IDI_FILE_FOOTPRINT_ICON 204
#define IDI_FILE_WORKSHEET_ICON 205
