// We extracted parts of the "stock" wx resource base
// because we want to follow Microsoft convention of numbering
// resources such as icons instead of playing alphabetical ordering games
// We also don't need the stock wx icons

// Standard wxWindows Cursors
WXCURSOR_HAND           CURSOR  DISCARDABLE     "wx/msw/hand.cur"
WXCURSOR_BULLSEYE       CURSOR  DISCARDABLE     "wx/msw/bullseye.cur"
WXCURSOR_PENCIL         CURSOR  DISCARDABLE     "wx/msw/pencil.cur"
WXCURSOR_MAGNIFIER      CURSOR  DISCARDABLE     "wx/msw/magnif1.cur"
WXCURSOR_ROLLER         CURSOR  DISCARDABLE     "wx/msw/roller.cur"
WXCURSOR_PBRUSH         CURSOR  DISCARDABLE     "wx/msw/pbrush.cur"
WXCURSOR_PLEFT          CURSOR  DISCARDABLE     "wx/msw/pntleft.cur"
WXCURSOR_PRIGHT         CURSOR  DISCARDABLE     "wx/msw/pntright.cur"
WXCURSOR_BLANK          CURSOR  DISCARDABLE     "wx/msw/blank.cur"
WXCURSOR_CROSS          CURSOR  DISCARDABLE     "wx/msw/cross.cur"

// For wxContextHelpButton
csquery                 BITMAP "wx/msw/csquery.bmp"

// For obtaining the RGB values of standard colours
wxBITMAP_STD_COLOURS    BITMAP "wx/msw/colours.bmp"

#include <winver.h> // defines VS_VERSION_INFO
#define INCLUDE_KICAD_VERSION
#include <kicad_build_version.h>

#define CREATEPROCESS_MANIFEST_RESOURCE_ID 1
#define ISOLATIONAWARE_MANIFEST_RESOURCE_ID 2
#define RT_MANIFEST 24

#ifdef ISOLATION_AWARE_ENABLED
#define MANIFEST_ID ISOLATIONAWARE_MANIFEST_RESOURCE_ID
#else
#define MANIFEST_ID CREATEPROCESS_MANIFEST_RESOURCE_ID
#endif

#if !defined( KICAD_WIN32_DPI_AWARE ) || KICAD_WIN32_DPI_AWARE == 0
#define MANIFEST_NAME basic.manifest
#else
#define MANIFEST_NAME dpi_aware_pmv2.manifest
#endif

//MSYS2
#define RC_STR( text ) RC_STR2( text )
#define RC_STR2( text ) #text
//MSVC
#define RC_CONCAT( a, b, c ) RC_CONCAT2( a, b, c )
#define RC_CONCAT2( a, b, c ) a##b##c

#ifdef __GNUC__
#define MANIFEST_FILE "manifests/" RC_STR( MANIFEST_NAME )
#else
//Do not try and quote the first part, it won't work, also make sure the IDE doesn't reformat it with spaces between slashes
#define MANIFEST_FILE RC_CONCAT( manifests/, MANIFEST_NAME )
#endif

MANIFEST_ID RT_MANIFEST MANIFEST_FILE


#ifndef DEBUG
#define VER_DEBUG 0
#else
#define VER_DEBUG VS_FF_DEBUG
#endif

VS_VERSION_INFO VERSIONINFO
FILEVERSION KICAD_WIN32_RC_FILEVER
PRODUCTVERSION KICAD_WIN32_RC_PRODVER
FILEFLAGSMASK VS_FFI_FILEFLAGSMASK
FILEFLAGS (VER_DEBUG)
FILEOS VOS__WINDOWS32
FILETYPE VFT_APP
FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4"
        BEGIN
            // this gets displayed in the windows 10 task manager
            VALUE "FileDescription", RC_VER_FILE_DESCRIPTION
            VALUE "FileVersion", KICAD_WIN32_RC_FILEVER_STR
            //original filename without extension
            VALUE "InternalName", RC_VER_INTERNALNAME
            VALUE "LegalCopyright", "Copyright (C) 2024 KiCad Developers"
            // the exe name
            VALUE "OriginalFilename", RC_VER_ORIGINALFILENAME
            // name of distributed package
            VALUE "ProductName", "KiCad EDA"
            VALUE "ProductVersion", KICAD_WIN32_RC_PRODVER_STR
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1252
    END
END

//RESOURCE FILES MUST CONTAIN A BLANK LINE AT THE END
