/*
 *
 * This file is part of KiCad, a free EDA CAD application.
 *
 * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef KICAD_SPACEMOUSE_H
#define KICAD_SPACEMOUSE_H
#include <wx/event.h>

// Define new events for SpaceMouse motion and button
wxDECLARE_EVENT(EVT_SPACEMOUSE_MOTION, wxCommandEvent);
wxDECLARE_EVENT(EVT_SPACEMOUSE_BUTTON, wxCommandEvent);

class EDA_DRAW_PANEL_GAL;

class LINUX_SPACEMOUSE
{
public:
    LINUX_SPACEMOUSE(EDA_DRAW_PANEL_GAL* parent);
    ~LINUX_SPACEMOUSE();

private:
    EDA_DRAW_PANEL_GAL* m_parent;
    void OnSpaceMouseMotion();
    void OnSpaceMouseButton();
};

#endif // KICAD_SPACEMOUSE_H