/*
 * This program source code file is part of KiCad, a free EDA CAD application.
 *
 * Copyright The KiCad Developers, see CHANGELOG.txt for contributors.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you may find one here:
 * http://www.gnu.org/licenses/gpl-3.0.html
 * or you may search the http://www.gnu.org website for the version 3 license,
 * or you may write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#ifndef PCBNEW_SPACENAV_2D_PLUGIN_H
#define PCBNEW_SPACENAV_2D_PLUGIN_H

#include <memory>
#include <wx/timer.h>

#include <kicommon.h>
#include <math/vector2d.h>
#include <spacenav/spacenav_driver.h>
#include <spacenav/libspnav_driver.h>

class EDA_DRAW_PANEL_GAL;
namespace KIGFX
{
    class VIEW;
}

class KICOMMON_API SPNAV_2D_PLUGIN : public wxEvtHandler, public SPACEMOUSE_HANDLER
{
public:
    explicit SPNAV_2D_PLUGIN( EDA_DRAW_PANEL_GAL* aCanvas );
    ~SPNAV_2D_PLUGIN();

    void SetFocus( bool aFocus = true );

    void SetCanvas( EDA_DRAW_PANEL_GAL* aCanvas );

    void SetScale( double aScale )
    {
        m_scale = aScale;
    }

    void OnPan( double x, double y, double z ) override;
    void OnRotate( double rx, double ry, double rz ) override;
    void OnButton( int button, bool pressed ) override;

private:
    void onPollTimer( wxTimerEvent& evt );

    std::unique_ptr<SPACENAV_DRIVER> m_driver;
    wxTimer                          m_timer;
    EDA_DRAW_PANEL_GAL*              m_canvas;
    KIGFX::VIEW*                     m_view;
    double                           m_scale;
    bool                             m_focused;
};

#endif // PCBNEW_SPACENAV_2D_PLUGIN_H