///////////////////////////////////////////////////////////////////////////
// C++ code generated with wxFormBuilder (version 4.2.1-0-g80c4cb6)
// http://www.wxformbuilder.org/
//
// PLEASE DO *NOT* EDIT THIS FILE!
///////////////////////////////////////////////////////////////////////////

#pragma once

#include <wx/artprov.h>
#include <wx/xrc/xmlres.h>
#include <wx/intl.h>
class STD_BITMAP_BUTTON;

#include "dialog_shim.h"
#include <wx/string.h>
#include <wx/stattext.h>
#include <wx/gdicmn.h>
#include <wx/font.h>
#include <wx/colour.h>
#include <wx/settings.h>
#include <wx/textctrl.h>
#include <wx/choice.h>
#include <wx/bmpbuttn.h>
#include <wx/bitmap.h>
#include <wx/image.h>
#include <wx/icon.h>
#include <wx/button.h>
#include <wx/sizer.h>
#include <wx/checklst.h>
#include <wx/dialog.h>

///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
/// Class DIALOG_DESTINATION_BASE
///////////////////////////////////////////////////////////////////////////////
class DIALOG_DESTINATION_BASE : public DIALOG_SHIM
{
	private:

	protected:
		wxStaticText* m_textArchiveDesc;
		wxTextCtrl* m_textCtrlDescription;
		wxStaticText* m_textArchiveFormat;
		wxChoice* m_choiceArchiveformat;
		wxStaticText* m_textOutputPath;
		wxTextCtrl* m_textCtrlOutputPath;
		STD_BITMAP_BUTTON* m_buttonOutputPath;
		wxStaticText* m_staticText10;
		wxCheckListBox* m_includeJobs;
		wxStdDialogButtonSizer* m_sdbSizer1;
		wxButton* m_sdbSizer1OK;
		wxButton* m_sdbSizer1Cancel;

		// Virtual event handlers, override them in your derived class
		virtual void onOutputPathBrowseClicked( wxCommandEvent& event ) { event.Skip(); }


	public:

		DIALOG_DESTINATION_BASE( wxWindow* parent, wxWindowID id = wxID_ANY, const wxString& title = _("%s Destination"), const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxSize( -1,-1 ), long style = wxDEFAULT_DIALOG_STYLE|wxRESIZE_BORDER );

		~DIALOG_DESTINATION_BASE();

};

