// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "If you specify the maximum current, then the track widths will be calculated to suit.\n"
"\n"
"If you specify one of the track widths, the maximum current it can handle will be calculated. The width for the other track to also handle this current will then be calculated.\n"
"\n"
"The controlling value is shown in bold.\n"
"\n"
"The calculations are valid for currents up to 35 A (external) or 17.5 A (internal), temperature rises up to 100 °C, and widths of up to 400 mils (10 mm).\n"
"\n"
"The formula, from IPC 2221, is\n"
"<center>___I = K &middot; &Delta;T<sup>0.44</sup> &middot; (W &middot; H)<sup>0.725</sup>___</center>\n"
"where:<br>\n"
"___I___ is maximum current in A<br>\n"
"___&Delta;T___ is temperature rise above ambient in &deg;C<br>\n"
"___W___ is width in mils<br>\n"
"___H___ is thickness (height) in mils<br>\n"
"___K___ is 0.024 for internal tracks or 0.048 for external tracks\n"
"" );
