/*
 * This program source code file is part of KiCad, a free EDA CAD application.
 *
 * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
 * Author: SYSUEric <jzzhuang666@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef _ODB_DEFINES_H_
#define _ODB_DEFINES_H_

#include <string>

#define ODB_JOB_NAME "JOB_NAME"
#define ODB_UNITS "UNITS"
#define ODB_DIM_X "x"
#define ODB_DIM_R "r"
#define ODB_DIM_C "c"
#define ODB_NONE "NONE"

#endif // _ODB_DEFINES_H_
