/*
 * This program source code file is part of KiCad, a free EDA CAD application.
 *
 * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you may find one here:
 * http://www.gnu.org/licenses/gpl-3.0.html
 * or you may search the http://www.gnu.org website for the version 3 license,
 * or you may write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#pragma once

#include <utility>
#include <vector>

#include <wx/listctrl.h>
#include <wx/string.h>

#include "dialog_autosave_recovery_base.h"


enum class AUTOSAVE_RECOVERY_CHOICE
{
    RESTORE,
    KEEP_CURRENT,
    KEEP_BOTH,
    CANCEL
};


class DIALOG_AUTOSAVE_RECOVERY : public DIALOG_AUTOSAVE_RECOVERY_BASE
{
public:
    DIALOG_AUTOSAVE_RECOVERY( wxWindow* aParent, const std::vector<std::pair<wxString, wxString>>& aStale );

    AUTOSAVE_RECOVERY_CHOICE GetChoice() const { return m_choice; }

    // Subset of the input stale list that the user kept checked.
    std::vector<std::pair<wxString, wxString>> GetSelectedStale() const;

protected:
    void OnRestore( wxCommandEvent& aEvent ) override;
    void OnKeepCurrent( wxCommandEvent& aEvent ) override;
    void OnKeepBoth( wxCommandEvent& aEvent ) override;
    void OnCancel( wxCommandEvent& aEvent ) override;

    void onItemCheckChanged( wxListEvent& aEvent );
    void updateActionButtons();

private:
    AUTOSAVE_RECOVERY_CHOICE                   m_choice;
    std::vector<std::pair<wxString, wxString>> m_stale;
};
