// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "SPICE functions:\n"
"\n"
"    sqrt(x)\n"
"    sin(x)\n"
"    cos(x)\n"
"    tan(x)\n"
"    sinh(x)\n"
"    cosh(x)\n"
"    tanh(x)\n"
"    asin(x)\n"
"    acos(x)\n"
"    atan(x)\n"
"    asinh(x)\n"
"    acosh(x)\n"
"    atanh(x)\n"
"    arctan(x)\n"
"    exp(x)\n"
"    ln(x)\n"
"    log(x)\n"
"    abs(x)\n"
"    nint(x)\n"
"    int(x)\n"
"    floor(x)\n"
"    ceil(x)\n"
"    pow(x, y)\n"
"    pwr(x, y)\n"
"    min(x, y)\n"
"    max(x, y)\n"
"    sgn(x)\n"
"    ternary_fcn(x, y, z)\n"
"    gauss(nom, rvar, sigma)\n"
"    agauss(nom, avar, sigma)\n"
"    unif(nom, rvar)\n"
"    aunif(nom, avar)\n"
"    limit(nom, avar)\n"
"" );
