// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "### Properties\n"
"\n"
"Many properties of objects are available for testing in custom rule expressions. See the full documentation at [https://docs.kicad.org](https://go.kicad.org/docs/GetMajorMinorVersion/en/pcbnew/pcbnew.html#custom_design_rules) for the complete list.\n"
"<br><br><br>\n"
"\n"
"" );
