/*
* This program source code file is part of KiCad, a free EDA CAD application.
*
* Copyright The KiCad Developers, see AUTHORS.txt for contributors.
*
* This program is free software: you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SYMBOL_SAVEAS_TYPE_H
#define SYMBOL_SAVEAS_TYPE_H

enum class SYMBOL_SAVEAS_TYPE
{
    NORMAL_SAVE_AS = 0,
    REPLACE_TABLE_ENTRY,
    ADD_GLOBAL_TABLE_ENTRY,
    ADD_PROJECT_TABLE_ENTRY
};

#endif