#define RC_VER_FILE_DESCRIPTION "KiCad Drawing Sheet Editor " KICAD_WIN32_RC_PRODVER_STR
#define RC_VER_INTERNALNAME "_pl_editor"
#define RC_VER_ORIGINALFILENAME "_pl_editor.dll"
#include "kiwin32-dll.rc"
