// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "The goal of the IEC60664-1 is to provide guidance on designing insulation for products that have a connection to mains supply.\n"
"\n"
"However some cases are not covered by this calculator:\n"
"\n"
"- For frequencies higher than 30kHz, the dielectric performances are degraded. IEC60664-4 covers those cases\n"
"- When using a conformal coating or a potting in order to protect for pollution, if all conditions specified by IEC60664-3 are met, the clearance and creepage distances can be reduced. Soldermask is usually not considered as a conformal coating.\n"
"- Insulations trough liquids, compressed air or gases other than air are not in the scope of IEC60664" );
